private import std.ctype;
private import std.file;
private import std.string;

/**
 * INIファイル読み込みクラス
 */
class ConfigParser
{
private:
	int            _index;    // 現在のデータ番号
	char[][]       _indices;  // データ番号
	char[][char[]] _data;     // データ(キーと値)
	char[]         _filepath; // ファイルパス
public:
	this() {}
	void init()
	{
		// 初期化
		_index          = 0;
		_indices.length = 0;
		_data           = null;
	}
	void load(char[] filepath)
	{
		init();
		_filepath = filepath.dup; // ファイル名コピー
		// 読み込み
		char[] file;
		try
		{
			file = cast(char[])read(filepath);
		}
		catch(Exception e)
		{
			throw new Error(format("can't open file '%s'", filepath));
		}
		// データを入れる
		foreach(b; split(file, "\n"))
		{
			int idx = find(b, '#');       // コメント文字を検索
			if(idx >= 0) b = b[0..idx];   // コメントを除去
			char[][] tmp = split(b, "="); // キーと値で分割
			if(tmp is null)    continue;
			if(tmp.length < 2) continue;
			char[] key = strip(tmp[0]);
			char[] val = strip(tmp[1]);
			set(key, val);
		}
		_data.rehash;
	}
	char[] get(char[] key)        { checkKey(key);         return _data[key];             } // 文字列で取得
	int    getint(char[] key)     { char[] val = get(key); return checkInt(key, val);     } // 整数で取得
	float  getfloat(char[] key)   { char[] val = get(key); return checkFloat(key, val);   } // 浮動小数で取得
	bool   getboolean(char[] key) { char[] val = get(key); return checkBoolean(key, val); } // 真偽で取得
	void   set(char[] key, char[] val)
	{
		if(!(key in _data))
		{
			_index++;
			_indices ~= key;
		}
		_data[key] = val;
	}
		// 文字列の追加
	void   set(char[] key, int    val) { set(key, std.string.toString(val)); } // 整数の追加
	void   set(char[] key, float  val) { set(key, std.string.toString(val)); } // 浮動小数の追加
	void   set(char[] key, bool   val) { set(key, std.string.toString(val)); } // 真偽の追加
	void write(char[] filepath=null)
	{
		if(filepath is null) filepath = _filepath;
		std.file.write(filepath, "");
		foreach(k; _indices)
		{
			append(filepath, format("%s = %s\n", k, _data[k]));
		}
	}
	void dump()
	{
		printf("{\n");
		foreach(k; _data.keys)
		{
			printf("  \"%.*s\" : \"%.*s\",\n", k, _data[k]);
		}
		printf("}\n");
	}
	void dump2()
	{
		foreach(k; _indices)
		{
			printf("%.*s = %.*s\n", k, _data[k]);
		}
	}
private:
	void checkKey(char[] key)
	{
		if(key in _data) return;
		throw new Error(format("ConfigParser.checkKey(): not exist key ... '%s'", key));
	}
	int checkInt(char[] key, char[] val)
	{
		foreach(c; val)
		{
			if(!isdigit(c) && c != '-') throw new Error(format("not digit value ... '%s' -> '%s'", key, val));
		}
		return atoi(val);
	}
	float checkFloat(char[] key, char[] val)
	{
		foreach(c; val)
		{
			if(!isdigit(c) && c != '.') throw new Error(format("not digit value ... '%s' -> '%s'", key, val));
		}
		return atof(val);
	}
	bool checkBoolean(char[] key, char[] val)
	{
		switch(val)
		{
		case "1", "yes", "true",  "on":  return true;
		case "0", "no",  "false", "off": return false;
		default: throw new Error(format("not boolean value ... '%s' -> '%s'", key, val));
		}
	}
}
